"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.IcExecutionMachine = exports.Scope = void 0;
const assignment_1 = require("./ic/assignment");
const atom_1 = require("./ic/atom");
const binary_operation_1 = require("./ic/binary-operation");
const break_1 = require("./ic/break");
const continue_1 = require("./ic/continue");
const declaration_statement_1 = require("./ic/declaration-statement");
const do_while_1 = require("./ic/do-while");
const expression_statement_1 = require("./ic/expression-statement");
const function_1 = require("./ic/function");
const function_call_1 = require("./ic/function-call");
const if_1 = require("./ic/if");
const return_1 = require("./ic/return");
const statements_1 = require("./ic/statements");
const variable_declaration_1 = require("./ic/variable-declaration");
const variable_reference_1 = require("./ic/variable-reference");
const while_1 = require("./ic/while");
class Scope {
    constructor(options) {
        this.parent = options?.parent;
        this.variables = new Map();
        this.functions = new Map();
    }
    create_child() {
        return new Scope({ parent: this });
    }
    declare_variable(name, type, value) {
        if (this.variables.has(name))
            throw new Error(`duplicate variable \`${name}\` declaration`);
        this.variables.set(name, { type, value });
    }
    get_variable(name) {
        const variable = this.variables.get(name);
        if (variable === undefined) {
            if (this.parent === undefined) {
                throw new Error(`variable \`${name}\` is not declared`);
            }
            else {
                return this.parent.get_variable(name);
            }
        }
        return variable;
    }
    declare_function(name, f) {
        if (this.functions.has(name))
            throw new Error(`duplicate  function \`${name}\` declaration`);
        this.functions.set(name, f);
    }
    get_function(name) {
        const f = this.functions.get(name);
        if (f === undefined) {
            if (this.parent === undefined) {
                throw new Error(`function \`${name}\` is not declared`);
            }
            else {
                return this.parent.get_function(name);
            }
        }
        return f;
    }
}
exports.Scope = Scope;
class IcExecutionMachine {
    constructor() {
        this.global_scope = new Scope();
    }
    evaluate(e, options) {
        const scope = options?.scope || this.global_scope;
        if (e instanceof atom_1.Atom) {
            return e.value;
        }
        else if (e instanceof binary_operation_1.BinaryOperation) {
            const left = this.evaluate(e.left_operand, { scope });
            switch (e.operator) {
                case binary_operation_1.Operator.OR: return left || this.evaluate(e.right_operand, { scope });
                case binary_operation_1.Operator.AND: return left && this.evaluate(e.right_operand, { scope });
            }
            const right = this.evaluate(e.right_operand, { scope });
            switch (e.operator) {
                case binary_operation_1.Operator.PLUS: return left + right;
                case binary_operation_1.Operator.MINUS: return left - right;
                case binary_operation_1.Operator.MULTIPLY: return left * right;
                case binary_operation_1.Operator.DIVIDE: return Math.floor(left / right);
                case binary_operation_1.Operator.EQUAL: return left === right;
                case binary_operation_1.Operator.NOT_EQUAL: return left !== right;
                case binary_operation_1.Operator.LESS_THAN: return left < right;
                case binary_operation_1.Operator.LESS_THAN_OR_EQUAL: return left <= right;
                case binary_operation_1.Operator.GREATER_THAN: return left > right;
                case binary_operation_1.Operator.GREATER_THAN_OR_EQUAL: return left >= right;
                default:
                    throw new Error(`unknown operator: ${binary_operation_1.Operator[e.operator]}`);
            }
        }
        else if (e instanceof function_call_1.FunctionCall) {
            const f = scope.get_function(e.name);
            const args = e.args.map(x => this.evaluate(x, { scope }));
            return f(...args);
        }
        else if (e instanceof variable_reference_1.VariableReference) {
            return scope.get_variable(e.target.name).value;
        }
        else {
            throw new Error('lack of implementation');
        }
    }
    execute(s, options) {
        const scope = options?.scope || this.global_scope;
        if (s instanceof assignment_1.Assignment) {
            const variable = scope.get_variable(s.variable_reference.target.name);
            variable.value = this.evaluate(s.expression, { scope });
        }
        else if (s instanceof break_1.Break) {
            return { type: 'break' };
        }
        else if (s instanceof continue_1.Continue) {
            return { type: 'continue' };
        }
        else if (s instanceof declaration_statement_1.DeclarationStatement) {
            const d = s.declaration;
            if (d instanceof function_1.Function) {
                scope.declare_function(d.name, (...args) => {
                    const function_scope = scope.create_child();
                    for (let i = 0; i < d.params.length; ++i) {
                        const param = d.params[i];
                        function_scope.declare_variable(param.name, param.type, args[i]);
                    }
                    const result = this.execute(d.body, { scope: function_scope });
                    if (d.return_type !== variable_declaration_1.VariableType.VOID) {
                        if (result?.type !== 'return')
                            throw new Error(`executed function \`${d.name}\` didn't returned any value`);
                        return result.value;
                    }
                });
            }
            else if (d instanceof variable_declaration_1.VariableDeclaration) {
                const initial_value = d.initial_value !== undefined
                    ? this.evaluate(d.initial_value, { scope })
                    : new atom_1.Atom(0);
                scope.declare_variable(d.name, d.type, initial_value);
            }
            else {
                throw new Error('lack of implementation');
            }
        }
        else if (s instanceof if_1.If) {
            if (this.evaluate(s.condition, { scope })) {
                const body_scope = scope.create_child();
                return this.execute(s.body, { scope: body_scope });
            }
            else {
                if (s.else_body !== undefined) {
                    const else_body_scope = scope.create_child();
                    return this.execute(s.else_body, { scope: else_body_scope });
                }
            }
        }
        else if (s instanceof expression_statement_1.ExpressionStatement) {
            this.evaluate(s.expression, { scope });
        }
        else if (s instanceof return_1.Return) {
            return {
                type: 'return',
                value: this.evaluate(s.expression, { scope })
            };
        }
        else if (s instanceof statements_1.Statements) {
            for (const statement of s.statements) {
                const value = this.execute(statement, { scope });
                if (value !== undefined)
                    return value;
            }
        }
        else if (s instanceof while_1.While) {
            while (this.evaluate(s.condition, { scope })) {
                const while_scope = scope.create_child();
                const value = this.execute(s.body, { scope: while_scope });
                if (value !== undefined) {
                    if (value.type === 'continue')
                        continue;
                    if (value.type === 'break')
                        break;
                    return value;
                }
            }
        }
        else if (s instanceof do_while_1.DoWhile) {
            do {
                const while_scope = scope.create_child();
                const value = this.execute(s.body, { scope: while_scope });
                if (value !== undefined) {
                    if (value.type === 'continue')
                        continue;
                    if (value.type === 'break')
                        break;
                    return value;
                }
            } while (this.evaluate(s.condition, { scope }));
        }
        else {
            throw new Error('lack of implementation');
        }
        return undefined;
    }
}
exports.IcExecutionMachine = IcExecutionMachine;
//# sourceMappingURL=ic-execution-machine.js.map