"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.generate = void 0;
const declaration_1 = require("./ast/declaration");
const atom_1 = require("./ic/atom");
const binary_operation_1 = require("./ic/binary-operation");
const declaration_statement_1 = require("./ic/declaration-statement");
const function_1 = require("./ic/function");
const function_call_1 = require("./ic/function-call");
const if_1 = require("./ic/if");
const return_1 = require("./ic/return");
const statements_1 = require("./ic/statements");
const variable_declaration_1 = require("./ic/variable-declaration");
const regex_to_ic_1 = require("./regex-to-ic");
const regular_expression_1 = require("./regular-expression");
class ICGenerator {
    constructor(regex_context, ast) {
        this.regex_context = regex_context;
        this.variables_map = new Map();
        this.terminals_map = new Map();
        this.terminals_parsing_functions_map = new Map();
        for (const declaration of ast.declarations) {
            switch (declaration.type) {
                case declaration_1.DeclarationType.DELIMITER: {
                    this.delimiter = declaration;
                    break;
                }
                case declaration_1.DeclarationType.TERMINAL: {
                    const id = this.terminals_map.size + 1;
                    const terminal = Object.assign(declaration, { id });
                    terminal.regex.machine_id = id;
                    this.terminals_map.set(declaration.name, terminal);
                    break;
                }
                case declaration_1.DeclarationType.VARIABLE: {
                    this.variables_map.set(declaration.name, declaration);
                    break;
                }
            }
        }
    }
    generate_function_for_parsing_terminals(terminals, options) {
        const epsilon = options?.epsilon === true;
        const terminals_names = terminals.map(x => x.name).sort();
        const name = `parse${epsilon ? '_eps' : ''}_${terminals_names.map(x => `T${x}`).join('_')}`;
        let f = this.terminals_parsing_functions_map.get(name);
        if (f === undefined) {
            const parsed_terminal_id_var = new variable_declaration_1.VariableDeclaration(variable_declaration_1.VariableType.I32, 'parsed_terminal_id', {
                mutable: true,
                initial_value: new atom_1.Atom(-1),
                comment: 'Contains the id of the last parsed terminal,\nor -1 if there was no terminal successfully parsed.'
            });
            const parsed_terminal_id_ref = parsed_terminal_id_var.get_reference();
            const machine = regular_expression_1.RegularExpression.merge(this.regex_context, terminals.map(x => x.regex));
            const statements = [
                parsed_terminal_id_var.to_statement(),
                (0, regex_to_ic_1.regex_to_ic)(machine, { parsed_terminal_id_ref })
            ];
            if (!epsilon) {
                const error_message = (terminals.length === 1
                    ? `expected terminal \`${terminals[0].name}\`, `
                    : `expected one of terminals: ${terminals.map(x => `\`${x.name}\``).join(', ')}; `) + `but found "{}"`;
                const statement = new if_1.If(new binary_operation_1.BinaryOperation(binary_operation_1.Operator.EQUAL, parsed_terminal_id_ref, new atom_1.Atom(-1)), (new function_call_1.FunctionCall('throw_error', { args: [
                        new atom_1.Atom(error_message),
                        new function_call_1.FunctionCall('read')
                    ] })).to_statement());
                statements.push(statement);
            }
            statements.push(new return_1.Return(parsed_terminal_id_ref));
            f = new function_1.Function(name, [], variable_declaration_1.VariableType.I32, new statements_1.Statements(statements), {
                comment: `Parses one of terminals: ${terminals_names.map(x => `\`${x}\``).join(', ')}.`
            });
            this.terminals_parsing_functions_map.set(name, f);
        }
        return f;
    }
    generate_function_consume_delimiter(delimiter) {
        return new function_1.Function('consume_delimiter', [], variable_declaration_1.VariableType.VOID, new statements_1.Statements([
            (0, regex_to_ic_1.regex_to_ic)(delimiter.regex),
            (new function_call_1.FunctionCall('reset')).to_statement()
        ]));
    }
    generate_function_for_parsing_variable(variable) {
        const all_possible_terminals = variable.terminals_maps.map(m => m.terminals).flat();
        const parsing_function = this.generate_function_for_parsing_terminals(all_possible_terminals, { epsilon: variable.epsilon });
        const variable_terminal_id = new variable_declaration_1.VariableDeclaration(variable_declaration_1.VariableType.I32, 'terminal_id', {
            initial_value: new function_call_1.FunctionCall(parsing_function.name)
        });
        let branches = !variable.epsilon
            ? (new function_call_1.FunctionCall('throw_error', { args: [
                    new atom_1.Atom('program reached invalid state: unexpected terminal (id: {}) parsed'),
                    variable_terminal_id.get_reference()
                ] })).to_statement()
            : undefined;
        for (const terminals_map of variable.terminals_maps) {
            const statements = [];
            for (const node of terminals_map.nodes.slice(1)) {
                if (node.name.charAt(0) == node.name.charAt(0).toLocaleLowerCase()) {
                    const terminal_parsing_function = this.generate_function_for_parsing_terminals([node.reference]);
                    if (this.delimiter !== undefined) {
                        statements.push((new function_call_1.FunctionCall('consume_delimiter')).to_statement());
                    }
                    statements.push((new function_call_1.FunctionCall(terminal_parsing_function.name)).to_statement(), (new function_call_1.FunctionCall('reset')).to_statement());
                }
                else {
                    statements.push((new function_call_1.FunctionCall(`parse_V${node.name}`)).to_statement());
                }
            }
            const condition = terminals_map.terminals.map(x => new binary_operation_1.BinaryOperation(binary_operation_1.Operator.EQUAL, variable_terminal_id.get_reference(), new atom_1.Atom(this.terminals_map.get(x.name).id))).reduce((left, right) => new binary_operation_1.BinaryOperation(binary_operation_1.Operator.OR, left, right));
            branches = new if_1.If(condition, new statements_1.Statements(statements), {
                else_body: branches
            });
        }
        if (branches === undefined)
            throw new Error('creating variable parsing function without branches');
        const body = new statements_1.Statements([]);
        if (this.delimiter !== undefined) {
            body.statements.push((new function_call_1.FunctionCall('consume_delimiter')).to_statement());
        }
        body.statements.push(new declaration_statement_1.DeclarationStatement(variable_terminal_id, {
            comment: 'The id of the currently parsed terminal.'
        }), (new function_call_1.FunctionCall('reset')).to_statement(), branches);
        return new function_1.Function(`parse_V${variable.name}`, [], variable_declaration_1.VariableType.VOID, body, {
            comment: `Function that parses the \`${variable.name}\` variable.`
        });
    }
    generate_function_root(root_variable_parsing_function) {
        const var_last_char = new variable_declaration_1.VariableDeclaration(variable_declaration_1.VariableType.I32, 'last_char', {
            initial_value: new function_call_1.FunctionCall('next')
        });
        return new function_1.Function('root', [], variable_declaration_1.VariableType.VOID, new statements_1.Statements([
            (new function_call_1.FunctionCall(root_variable_parsing_function.name)).to_statement(),
            (new function_call_1.FunctionCall('reset')).to_statement(),
            ...((() => {
                return this.delimiter !== undefined
                    ? [(new function_call_1.FunctionCall('consume_delimiter')).to_statement()]
                    : [];
            })()),
            var_last_char.to_statement(),
            new if_1.If(new binary_operation_1.BinaryOperation(binary_operation_1.Operator.NOT_EQUAL, var_last_char.get_reference(), new atom_1.Atom(0)), (new function_call_1.FunctionCall('throw_error', {
                args: [
                    new atom_1.Atom('EOF expected, but something else found')
                ]
            })).to_statement())
        ]));
    }
    generate() {
        const declarations = [];
        if (this.delimiter !== undefined) {
            declarations.push(this.generate_function_consume_delimiter(this.delimiter));
        }
        const variables_parsing_functions = Array.from(this.variables_map.values()).map(x => this.generate_function_for_parsing_variable(x));
        declarations.push(...this.terminals_parsing_functions_map.values(), ...variables_parsing_functions, this.generate_function_root(variables_parsing_functions[variables_parsing_functions.length - 1]));
        return declarations;
    }
}
function generate(regex_context, ast) {
    const generator = new ICGenerator(regex_context, ast);
    return generator.generate();
}
exports.generate = generate;
//# sourceMappingURL=intermediate-code-generation.js.map