"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.parse = exports.SymbolType = void 0;
const location_1 = require("./source/location");
const error_1 = require("./error");
var SymbolType;
(function (SymbolType) {
    SymbolType[SymbolType["IDENTIFIER"] = 0] = "IDENTIFIER";
    SymbolType[SymbolType["DELIMITER"] = 1] = "DELIMITER";
    SymbolType[SymbolType["TERMINAL"] = 2] = "TERMINAL";
    SymbolType[SymbolType["REGEX"] = 3] = "REGEX";
    SymbolType[SymbolType["VARIABLE"] = 4] = "VARIABLE";
    SymbolType[SymbolType["PRODUCTION"] = 5] = "PRODUCTION";
    SymbolType[SymbolType["EPSILON"] = 6] = "EPSILON";
    SymbolType[SymbolType["BRACES_LEFT"] = 7] = "BRACES_LEFT";
    SymbolType[SymbolType["BRACES_RIGHT"] = 8] = "BRACES_RIGHT";
    SymbolType[SymbolType["EOF"] = 9] = "EOF";
})(SymbolType || (exports.SymbolType = SymbolType = {}));
const REGEX_ID_START = /^\p{L}$/u;
const REGEX_ID_CONTINUE = /^[\p{L}\d_]$/u;
function parse(reader) {
    const symbols = [];
    while (true) {
        let lexeme = reader.read();
        let location = reader.get_point();
        let type;
        if (lexeme === '')
            break;
        if (/^\s$/.test(lexeme)) {
            continue;
        }
        else if (lexeme === '{') {
            type = SymbolType.BRACES_LEFT;
        }
        else if (lexeme === '}') {
            type = SymbolType.BRACES_RIGHT;
        }
        else if (lexeme === '/') {
            type = SymbolType.REGEX;
            let escaping = false;
            while (true) {
                const c = reader.read();
                lexeme += c;
                if (escaping) {
                    escaping = false;
                }
                else {
                    if (c === '\\') {
                        escaping = true;
                    }
                    else if (c === '/') {
                        location = new location_1.Location(location, reader.get_point());
                        break;
                    }
                }
            }
        }
        else if (REGEX_ID_START.test(lexeme)) {
            let last_point;
            while (true) {
                const c = reader.read();
                if (!REGEX_ID_CONTINUE.test(c)) {
                    reader.revoke();
                    break;
                }
                else {
                    lexeme += c;
                    last_point = reader.get_point();
                }
            }
            if (last_point !== undefined)
                location = new location_1.Location(location, last_point);
            switch (lexeme) {
                case 'delimiter': {
                    type = SymbolType.DELIMITER;
                    break;
                }
                case 'terminal': {
                    type = SymbolType.TERMINAL;
                    break;
                }
                case 'variable': {
                    type = SymbolType.VARIABLE;
                    break;
                }
                case 'production': {
                    type = SymbolType.PRODUCTION;
                    break;
                }
                case 'epsilon': {
                    type = SymbolType.EPSILON;
                    break;
                }
                default: {
                    type = SymbolType.IDENTIFIER;
                    break;
                }
            }
        }
        else {
            throw new error_1.CodeError(reader.file, location, `Unknown lexeme \`${lexeme}\``);
        }
        symbols.push({
            type,
            location,
            lexeme
        });
    }
    symbols.push({
        type: SymbolType.EOF,
        location: reader.get_point(),
        lexeme: ''
    });
    return symbols;
}
exports.parse = parse;
//# sourceMappingURL=lexical-analysis.js.map