"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.TypescriptOutputGenerator = void 0;
const assignment_1 = require("../ic/assignment");
const atom_1 = require("../ic/atom");
const binary_operation_1 = require("../ic/binary-operation");
const break_1 = require("../ic/break");
const continue_1 = require("../ic/continue");
const declaration_statement_1 = require("../ic/declaration-statement");
const do_while_1 = require("../ic/do-while");
const expression_statement_1 = require("../ic/expression-statement");
const function_1 = require("../ic/function");
const function_call_1 = require("../ic/function-call");
const if_1 = require("../ic/if");
const return_1 = require("../ic/return");
const statements_1 = require("../ic/statements");
const variable_declaration_1 = require("../ic/variable-declaration");
const variable_reference_1 = require("../ic/variable-reference");
const while_1 = require("../ic/while");
const code_builder_1 = require("../utils/code-builder");
function operator_to_priority(operator) {
    switch (operator) {
        case binary_operation_1.Operator.OR:
            return 1;
        case binary_operation_1.Operator.AND:
            return 2;
        case binary_operation_1.Operator.EQUAL:
        case binary_operation_1.Operator.NOT_EQUAL:
            return 3;
        case binary_operation_1.Operator.LESS_THAN:
        case binary_operation_1.Operator.LESS_THAN_OR_EQUAL:
        case binary_operation_1.Operator.GREATER_THAN:
        case binary_operation_1.Operator.GREATER_THAN_OR_EQUAL:
            return 4;
        case binary_operation_1.Operator.PLUS:
        case binary_operation_1.Operator.MINUS:
            return 5;
        case binary_operation_1.Operator.MULTIPLY:
        case binary_operation_1.Operator.DIVIDE:
            return 6;
        default:
            throw new Error('unknown operator');
    }
}
class TypescriptOutputGenerator {
    value_to_literal(value) {
        switch (typeof value) {
            case 'boolean':
                return Boolean(value).toString();
            case 'number':
                return Number(value).toString();
            case 'string':
                return JSON.stringify(value);
            default:
                throw new Error(`can't convert value \`${value}\` to literal`);
        }
    }
    ic_type_to_ts(type) {
        switch (type) {
            case variable_declaration_1.VariableType.VOID:
                return 'void';
            case variable_declaration_1.VariableType.I32:
                return 'number';
            default:
                throw new Error(`unknown variable type \`${variable_declaration_1.VariableType[type]}\``);
        }
    }
    generate_documentation_comment(builder, comment) {
        comment = comment?.trim();
        if (!!comment) {
            builder.println('/**');
            comment.split('\n').map(x => ` * ${x}`).forEach(x => builder.println(x));
            builder.println(' */');
        }
    }
    generate_expression(e, options) {
        let s;
        if (e instanceof atom_1.Atom) {
            s = this.value_to_literal(e.value);
        }
        else if (e instanceof variable_reference_1.VariableReference) {
            s = e.target.name;
        }
        else if (e instanceof binary_operation_1.BinaryOperation) {
            let operator;
            switch (e.operator) {
                case binary_operation_1.Operator.PLUS: {
                    operator = '+';
                    break;
                }
                case binary_operation_1.Operator.MINUS: {
                    operator = '-';
                    break;
                }
                case binary_operation_1.Operator.MULTIPLY: {
                    operator = '*';
                    break;
                }
                case binary_operation_1.Operator.DIVIDE: {
                    operator = '/';
                    break;
                }
                case binary_operation_1.Operator.OR: {
                    operator = '||';
                    break;
                }
                case binary_operation_1.Operator.AND: {
                    operator = '&&';
                    break;
                }
                case binary_operation_1.Operator.EQUAL: {
                    operator = '===';
                    break;
                }
                case binary_operation_1.Operator.NOT_EQUAL: {
                    operator = '!==';
                    break;
                }
                case binary_operation_1.Operator.LESS_THAN: {
                    operator = '<';
                    break;
                }
                case binary_operation_1.Operator.LESS_THAN_OR_EQUAL: {
                    operator = '<=';
                    break;
                }
                case binary_operation_1.Operator.GREATER_THAN: {
                    operator = '>';
                    break;
                }
                case binary_operation_1.Operator.GREATER_THAN_OR_EQUAL: {
                    operator = '>=';
                    break;
                }
                default:
                    throw new Error(`unknown operator: ${binary_operation_1.Operator[e.operator]}`);
            }
            const priority = operator_to_priority(e.operator);
            const left = this.generate_expression(e.left_operand, { parent_priority: priority });
            const right = this.generate_expression(e.right_operand, { parent_priority: priority });
            s = `${left} ${operator} ${right}`;
            if ((options?.parent_priority || -1) > priority) {
                s = `(${s})`;
            }
        }
        else if (e instanceof function_call_1.FunctionCall) {
            const args_expressions = e.args.map(x => this.generate_expression(x));
            if (e.name === 'throw_error') {
                const params = args_expressions.slice(1);
                const format = args_expressions[0];
                if (typeof format !== 'string')
                    throw new Error('format string passed to the `throw_error` function is not a string');
                s = `throw new Error(${format.replace(/\{\}/g, () => `\" + ${params.shift()} + \"`)})`;
            }
            else {
                s = `${e.name}(${args_expressions.join(', ')})`;
            }
        }
        else {
            throw new Error('lack of implementation: don\'t know how to generate expression');
        }
        return (e.comment !== undefined)
            ? `${s} /* ${e.comment} */`
            : s;
    }
    generate_statement(builder, s) {
        if (s.comment !== undefined) {
            builder.println(`// ${s.comment}`);
        }
        if (s instanceof statements_1.Statements) {
            s.statements.forEach(x => this.generate_statement(builder, x));
        }
        else if (s instanceof if_1.If) {
            builder
                .println(`if (${this.generate_expression(s.condition)}) {`)
                .wrap_indentation(() => {
                this.generate_statement(builder, s.body);
            });
            const else_body = s.else_body;
            if (else_body === undefined) {
                builder.println('}');
            }
            else {
                if (else_body instanceof if_1.If) {
                    builder.print('} else ');
                    this.generate_statement(builder, else_body);
                }
                else {
                    builder
                        .println('} else {')
                        .wrap_indentation(() => {
                        this.generate_statement(builder, else_body);
                    })
                        .println('}');
                }
            }
        }
        else if (s instanceof while_1.While) {
            builder
                .println(`while (${this.generate_expression(s.condition)}) {`)
                .wrap_indentation(() => {
                this.generate_statement(builder, s.body);
            })
                .println('}');
        }
        else if (s instanceof do_while_1.DoWhile) {
            builder
                .println('do {')
                .wrap_indentation(() => {
                this.generate_statement(builder, s.body);
            })
                .println(`} while (${this.generate_expression(s.condition)})`);
        }
        else if (s instanceof assignment_1.Assignment) {
            builder.println(`${s.variable_reference.target.name} = ${this.generate_expression(s.expression)}`);
        }
        else if (s instanceof return_1.Return) {
            builder.println(`return ${this.generate_expression(s.expression)}`);
        }
        else if (s instanceof expression_statement_1.ExpressionStatement) {
            builder.println(this.generate_expression(s.expression));
        }
        else if (s instanceof declaration_statement_1.DeclarationStatement) {
            this.generate_declaration(builder, s.declaration);
        }
        else if (s instanceof break_1.Break) {
            builder.println('break');
        }
        else if (s instanceof continue_1.Continue) {
            builder.println('continue');
        }
    }
    generate_declaration(builder, d) {
        if (d instanceof variable_declaration_1.VariableDeclaration) {
            this.generate_documentation_comment(builder, d.comment);
            const initialization = d.initial_value !== undefined
                ? ` = ${this.generate_expression(d.initial_value)}`
                : '';
            builder.println(`${d.mutable ? 'let' : 'const'} ${d.name}: ${this.ic_type_to_ts(d.type)}${initialization}`);
        }
        else if (d instanceof function_1.Function) {
            this.generate_documentation_comment(builder, [
                d.comment || '',
                ...d.params.map(x => `@param ${x.name} ${x.comment}`)
            ].join('\n'));
            const params = d.params.map(x => `${x.name}: ${this.ic_type_to_ts(x.type)}`).join(', ');
            builder
                .println(`function ${d.name}(${params}): ${this.ic_type_to_ts(d.return_type)} {`)
                .wrap_indentation(() => {
                this.generate_statement(builder, d.body);
            })
                .println('}');
        }
        else {
            throw new Error('lack of implementation');
        }
    }
    generate(declarations) {
        const builder = new code_builder_1.CodeBuilder();
        builder
            .println('export function parse(source: Iterable<number>): void {')
            .wrap_indentation(() => {
            builder
                .println('const iterator = source[Symbol.iterator]()')
                .println('let buffer: number[] = []')
                .println('let cursor_index = -1')
                .println('let marked_length = -1')
                .newline();
            builder
                .println('function current(): number {').wrap_indentation(() => {
                builder
                    .println('if (cursor_index === -1)')
                    .wrap_indentation(() => {
                    builder.println(`throw new Error('trying to read from uninitialized parser')`);
                })
                    .newline()
                    .println('if (buffer.length === cursor_index) {')
                    .wrap_indentation(() => {
                    builder
                        .println('const result = iterator.next()')
                        .println('buffer.push(result.done ? 0 : result.value)');
                })
                    .println('}')
                    .newline()
                    .println('return buffer[cursor_index]');
            })
                .println('}')
                .newline();
            builder
                .println('function mark(offset?: number): void {').wrap_indentation(() => {
                builder.println('marked_length = cursor_index + 1 + (offset || 0)');
            })
                .println('}')
                .newline();
            builder
                .println('function next(): number {').wrap_indentation(() => {
                builder
                    .println('advance()')
                    .println('return current()');
            })
                .println('}')
                .newline();
            builder
                .println('function advance(): void {').wrap_indentation(() => {
                builder
                    .println('cursor_index += 1');
            })
                .println('}')
                .newline();
            builder
                .println('function reset(): void {').wrap_indentation(() => {
                builder
                    .println('buffer = buffer.slice(marked_length)')
                    .println('cursor_index = -1')
                    .println('marked_length = -1');
            })
                .println('}')
                .newline();
            builder
                .println('// @ts-ignore')
                .println('function lexeme(): string {').wrap_indentation(() => {
                builder.println('return buffer.slice(0, marked_length).map(x => String.fromCharCode(x)).join(\'\')');
            })
                .println('}')
                .newline();
            builder
                .println('// @ts-ignore')
                .println('function read(): string {').wrap_indentation(() => {
                builder.println('return buffer.slice(0, cursor_index + 1).map(x => String.fromCharCode(x)).join(\'\')');
            })
                .println('}')
                .newline();
            for (const declaration of declarations) {
                this.generate_declaration(builder, declaration);
                builder.ensure_empty_line();
            }
            const root_function = declarations[declarations.length - 1];
            builder.println(`return ${root_function.name}()`);
        })
            .println('}');
        return builder.to_string();
    }
}
exports.TypescriptOutputGenerator = TypescriptOutputGenerator;
//# sourceMappingURL=typescript-output-generator.js.map