"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const fs_1 = __importDefault(require("fs"));
const source_reader_1 = require("./source/source-reader");
const reader_1 = require("./reader");
const context_1 = require("./regex/context");
const lexical_analysis_1 = require("./lexical-analysis");
const syntax_analysis_1 = require("./syntax-analysis");
const semantic_analysis_1 = require("./semantic-analysis");
const intermediate_code_generation_1 = require("./intermediate-code-generation");
const output_generation_1 = require("./output-generation");
const typescript_output_generator_1 = require("./output-generators/typescript-output-generator");
const cplusplus_output_generator_1 = require("./output-generators/cplusplus-output-generator");
class Pipeline {
    constructor(args) {
        this.args = args;
    }
    prepare_output_generator(args) {
        const ext = args.output_path.split('.').pop()?.toLowerCase();
        switch (ext) {
            case 'ts':
                return new typescript_output_generator_1.TypescriptOutputGenerator();
            case 'cpp':
            case 'cxx':
            case 'c++':
            case 'hpp':
            case 'hxx':
            case 'h++':
                return new cplusplus_output_generator_1.CPlusPlusOutputGenerator();
            default:
                throw new Error(`don't know how to format a \`${ext}\` file`);
        }
    }
    run() {
        const output_generator = this.prepare_output_generator(this.args);
        const source = fs_1.default.readFileSync(this.args.file, 'utf8');
        const reader = new source_reader_1.SourceReader(new reader_1.StringReader(source), { file: this.args.file });
        const symbols = (0, lexical_analysis_1.parse)(reader);
        const regex_context = new context_1.Context();
        const ast = (0, syntax_analysis_1.parse)(regex_context, this.args.file, symbols);
        (0, semantic_analysis_1.analyze)(this.args.file, ast);
        const declarations = (0, intermediate_code_generation_1.generate)(regex_context, ast);
        const output = (0, output_generation_1.generate)(output_generator, declarations);
        fs_1.default.writeFileSync(this.args.output_path, output, 'utf8');
    }
}
exports.default = Pipeline;
//# sourceMappingURL=pipeline.js.map