"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.regex_to_ic = void 0;
const assignment_1 = require("./ic/assignment");
const atom_1 = require("./ic/atom");
const binary_operation_1 = require("./ic/binary-operation");
const break_1 = require("./ic/break");
const do_while_1 = require("./ic/do-while");
const function_call_1 = require("./ic/function-call");
const if_1 = require("./ic/if");
const statements_1 = require("./ic/statements");
const variable_declaration_1 = require("./ic/variable-declaration");
const while_1 = require("./ic/while");
function enhance_printable_comment(atom) {
    if (Number.isSafeInteger(atom.value))
        atom.comment = JSON.stringify(String.fromCharCode(atom.value));
    return atom;
}
function build_transition_condition(transition, input_expression) {
    const intervals = transition.symbol.set.get_intervals();
    const intervals_conditions = intervals.map(interval => (interval.length === 1)
        ? new binary_operation_1.BinaryOperation(binary_operation_1.Operator.EQUAL, input_expression, enhance_printable_comment(new atom_1.Atom(interval.start)))
        : new binary_operation_1.BinaryOperation(binary_operation_1.Operator.AND, new binary_operation_1.BinaryOperation(binary_operation_1.Operator.GREATER_THAN_OR_EQUAL, input_expression, enhance_printable_comment(new atom_1.Atom(interval.start))), new binary_operation_1.BinaryOperation(binary_operation_1.Operator.LESS_THAN_OR_EQUAL, input_expression, enhance_printable_comment(new atom_1.Atom(interval.end - 1)))));
    return intervals_conditions.reduce((left, right) => new binary_operation_1.BinaryOperation(binary_operation_1.Operator.OR, left, right));
}
function reduce_branches(branches, else_body) {
    const reducer = (else_body, branch) => {
        return new if_1.If(branch.condition, branch.body, { else_body });
    };
    return branches.reduceRight(reducer, else_body) || new statements_1.Statements([]);
}
class Vars {
    constructor() {
        this.input_var = undefined;
        this.state_var = undefined;
    }
    declare_state_var(initial_value) {
        if (this.state_var !== undefined)
            throw new Error('Variable `state` already declared.');
        this.state_var = new variable_declaration_1.VariableDeclaration(variable_declaration_1.VariableType.I32, 'state', {
            mutable: true,
            initial_value: new atom_1.Atom(initial_value),
            comment: 'Will contain the id of the current state of the machine.'
        });
        return this.state_var;
    }
    get input_ref() {
        if (this.input_var === undefined) {
            this.input_var = new variable_declaration_1.VariableDeclaration(variable_declaration_1.VariableType.I32, 'input', {
                mutable: true,
                initial_value: new atom_1.Atom(-1),
                comment: 'Will contain the current input of the machine.'
            });
        }
        return this.input_var.get_reference();
    }
    get state_ref() {
        if (this.state_var === undefined) {
            this.state_var = this.declare_state_var(-1);
        }
        return this.state_var.get_reference();
    }
    get_declarations() {
        return [this.input_var, this.state_var].filter((x) => x !== undefined);
    }
}
class Analysis {
    constructor(machine) {
        this.refs = new Map();
        for (const state of machine.get_transitively_reachable_states()) {
            for (const transition of state.transitions) {
                const value = this.get_refs(transition.state) + 1;
                this.refs.set(transition.state.id, value);
            }
        }
    }
    get_refs(state) {
        return this.refs.get(state.id) || 0;
    }
}
function regex_to_ic(machine, options) {
    const parsed_terminal_id_ref = options?.parsed_terminal_id_ref;
    const analysis = new Analysis(machine);
    const vars = new Vars();
    const processed = new Set();
    const defer_queue = [];
    function defer_state_handling(state) {
        if (!processed.has(state.id)) {
            processed.add(state.id);
            defer_queue.push(state);
        }
    }
    function handle_state(state, fail_body, options) {
        const seen = options?.seen || new Set();
        const root_state = options?.root_state || state;
        seen.add(state.id);
        processed.add(state.id);
        const statements = [];
        let handle_final = true;
        let init_input = false;
        let state_body;
        if (state.transitions.length === 0) {
            state_body = fail_body;
        }
        else {
            let input_expr;
            if (state.transitions.length === 1 && state.transitions[0].symbol.set.size === 1) {
                input_expr = new function_call_1.FunctionCall('next');
            }
            else {
                init_input = true;
                input_expr = vars.input_ref;
            }
            const transitions = {
                looping: [],
                jumping: []
            };
            for (const transition of state.transitions) {
                if (state.id === transition.state.id) {
                    transitions.looping.push(transition);
                }
                else {
                    transitions.jumping.push(transition);
                }
            }
            if (transitions.looping.length > 1)
                throw new Error('bug: no state should have more than one looping-transition');
            for (const transition of transitions.looping) {
                const condition = build_transition_condition(transition, input_expr);
                let loop;
                if (init_input) {
                    init_input = false;
                    loop = new do_while_1.DoWhile(condition, new assignment_1.Assignment(vars.input_ref, new function_call_1.FunctionCall('next')));
                }
                else {
                    loop = new while_1.While(condition, new statements_1.Statements([], { comment: 'nop' }));
                }
                statements.push(loop);
                if (state.is_final) {
                    statements.push((new function_call_1.FunctionCall('mark', { args: [new atom_1.Atom(-1)] })).to_statement());
                    if (parsed_terminal_id_ref !== undefined)
                        statements.push(new assignment_1.Assignment(parsed_terminal_id_ref, new atom_1.Atom(state.machine_id)));
                    handle_final = false;
                }
            }
            const branches = [];
            for (const transition of transitions.jumping) {
                let condition = build_transition_condition(transition, input_expr);
                let next = transition.state;
                let body;
                while (true) {
                    if (seen.has(next.id) || analysis.get_refs(next) >= 2) {
                        if (next.id === root_state.id) {
                            body = new statements_1.Statements([], { comment: `remains in state ${next.id}` });
                        }
                        else {
                            body = new assignment_1.Assignment(vars.state_ref, new atom_1.Atom(next.id));
                            defer_state_handling(next);
                        }
                        break;
                    }
                    if (!next.is_final && next.transitions.length === 1) {
                        const next_transition = next.transitions[0];
                        if (next_transition.symbol.set.size === 1) {
                            condition = new binary_operation_1.BinaryOperation(binary_operation_1.Operator.AND, condition, build_transition_condition(next_transition, new function_call_1.FunctionCall('next')));
                            next = next_transition.state;
                            continue;
                        }
                    }
                    body = handle_state(next, fail_body, { seen, root_state });
                    break;
                }
                branches.push({ condition, body });
            }
            state_body = reduce_branches(branches, fail_body);
        }
        if (handle_final && state.is_final) {
            statements.push((new function_call_1.FunctionCall('mark')).to_statement());
            if (parsed_terminal_id_ref !== undefined)
                statements.push(new assignment_1.Assignment(parsed_terminal_id_ref, new atom_1.Atom(state.machine_id)));
        }
        if (init_input)
            statements.push(new assignment_1.Assignment(vars.input_ref, new function_call_1.FunctionCall('next')));
        if (state_body !== undefined)
            statements.push(state_body);
        return new statements_1.Statements(statements);
    }
    const statements = [
        handle_state(machine, undefined)
    ];
    if (defer_queue.length !== 0) {
        const branches = [];
        while (true) {
            const state = defer_queue.shift();
            if (state === undefined)
                break;
            const condition = new binary_operation_1.BinaryOperation(binary_operation_1.Operator.EQUAL, vars.state_ref, new atom_1.Atom(state.id));
            const body = handle_state(state, new break_1.Break());
            branches.push({
                condition,
                body
            });
        }
        const fail_body = (new function_call_1.FunctionCall('throw_error', {
            args: [new atom_1.Atom('program reached invalid state: regex machine state unknown')]
        })).to_statement();
        const while_body = reduce_branches(branches, fail_body);
        statements.push(new if_1.If(new binary_operation_1.BinaryOperation(binary_operation_1.Operator.NOT_EQUAL, vars.state_ref, new atom_1.Atom(-1)), new while_1.While(new atom_1.Atom(true), while_body)));
    }
    return new statements_1.Statements([
        ...vars.get_declarations().map(x => x.to_statement()),
        ...statements
    ]);
}
exports.regex_to_ic = regex_to_ic;
//# sourceMappingURL=regex-to-ic.js.map