"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.State = void 0;
const abstract_symbol_1 = require("./abstract-symbol");
const transition_1 = require("./transition");
const non_deterministic_states_map_1 = require("./non-deterministic-states-map");
class State {
    constructor(id, options) {
        this.id = id;
        this.transitions = [];
        this.is_final = false;
        this.machine_id = undefined;
        options = options || {};
        this.is_final = options.is_final === true;
    }
    add_transitions(...transitions) {
        this.transitions.push(...transitions);
    }
    add_transition(symbol, state) {
        this.add_transitions(new transition_1.Transition(symbol, state));
    }
    add_epsilon_transition(state) {
        this.add_transitions(new transition_1.Transition(undefined, state));
    }
    remove_all_transitions() {
        this.transitions.length = 0;
    }
    get_reachable_transitions() {
        const reachable = [];
        const queue = [...this.transitions];
        const already_queued = new Set([...this.transitions]);
        while (true) {
            const transition = queue.shift();
            if (transition === undefined)
                break;
            if (transition.is_epsilon()) {
                for (let next_transition of transition.state.transitions) {
                    if (already_queued.has(next_transition))
                        continue;
                    queue.push(next_transition);
                    already_queued.add(next_transition);
                }
            }
            else {
                reachable.push(transition);
            }
        }
        return reachable;
    }
    get_transitively_reachable_states_iterable() {
        const state_this = this;
        return {
            [Symbol.iterator]() {
                const processed_states = new Map();
                const queue = [state_this];
                return {
                    next() {
                        while (true) {
                            const state = queue.shift();
                            if (state === undefined) {
                                return {
                                    done: true,
                                    value: undefined
                                };
                            }
                            if (processed_states.has(state.id))
                                continue;
                            processed_states.set(state.id, state);
                            queue.push(...state.transitions.map(transition => transition.state));
                            return {
                                done: false,
                                value: state
                            };
                        }
                    }
                };
            }
        };
    }
    get_transitively_reachable_states() {
        return Array.from(this.get_transitively_reachable_states_iterable());
    }
    get_transitively_reachable_final_states() {
        return this.get_transitively_reachable_states().filter(state => state.is_final);
    }
    expand_final_through_epsilon_transitions() {
        if (!this.is_final)
            return;
        const queue = [this];
        const already_queued = new Set([this.id]);
        while (true) {
            const state = queue.shift();
            if (state === undefined)
                break;
            for (let transition of state.transitions) {
                if (!transition.is_epsilon())
                    continue;
                const next_state = transition.state;
                if (already_queued.has(next_state.id))
                    continue;
                next_state.is_final = true;
                next_state.machine_id = this.machine_id;
                queue.push(next_state);
                already_queued.add(next_state.id);
            }
        }
    }
    become_final_through_epsilon_transitions() {
        const queue = [this];
        const already_queued = new Set([this.id]);
        while (true) {
            const state = queue.shift();
            if (state === undefined)
                break;
            if (state.is_final) {
                this.is_final = true;
                this.machine_id = state.machine_id;
                break;
            }
            for (let transition of state.transitions) {
                if (!transition.is_epsilon())
                    continue;
                const next_state = transition.state;
                if (already_queued.has(next_state.id))
                    continue;
                queue.push(next_state);
                already_queued.add(next_state.id);
            }
        }
    }
    get_transitions_multi_state_map() {
        const transitions = this.get_reachable_transitions();
        if (transitions.length === 0)
            return [];
        const map = [];
        const first_transition = transitions.shift();
        map.push({
            symbol: first_transition.symbol,
            states: [first_transition.state]
        });
        for (let transition of transitions) {
            let symbol = transition.symbol;
            for (let i = 0, n = map.length; i < n; i += 1) {
                if (!symbol.represents_something())
                    break;
                const entry = map[i];
                const fragmentation = abstract_symbol_1.AbstractSymbol.fragment(entry.symbol, symbol);
                if (fragmentation.shared.represents_something()) {
                    if (fragmentation.first_exclusive.represents_something()) {
                        entry.symbol = fragmentation.first_exclusive;
                        const states = [...entry.states];
                        if (states.find(x => x.id === transition.state.id) === undefined)
                            states.push(transition.state);
                        map.push({
                            symbol: fragmentation.shared,
                            states
                        });
                    }
                    else {
                        entry.symbol = fragmentation.shared;
                        if (entry.states.find(x => x.id === transition.state.id) === undefined)
                            entry.states.push(transition.state);
                    }
                }
                symbol = fragmentation.second_exclusive;
            }
            if (symbol.represents_something()) {
                map.push({
                    symbol,
                    states: [transition.state]
                });
            }
        }
        return map;
    }
    remove_non_determinism(context) {
        const all_states = this.get_transitively_reachable_states();
        const states_map = new non_deterministic_states_map_1.NonDeterministicStatesMap(context, all_states);
        const already_queued = new Set([this.id]);
        const queue = [this];
        while (true) {
            const state = queue.shift();
            if (state === undefined)
                break;
            state.become_final_through_epsilon_transitions();
            const map = state.get_transitions_multi_state_map();
            const transitions = [];
            for (let entry of map) {
                const next_state = states_map.get_or_create(entry.states);
                transitions.push(new transition_1.Transition(entry.symbol, next_state));
                if (!already_queued.has(next_state.id)) {
                    queue.push(next_state);
                    already_queued.add(next_state.id);
                }
            }
            state.remove_all_transitions();
            state.add_transitions(...transitions);
        }
    }
    match(input, options) {
        const symbols = Array.from(input);
        let last_match = undefined;
        let last_match_state = undefined;
        let index = 0;
        let state = this;
        while (true) {
            if (state.is_final) {
                last_match = index;
                last_match_state = state;
            }
            if (index === symbols.length)
                break;
            const code_point = symbols[index].codePointAt(0);
            const transition = state.transitions.find(x => x.symbol === undefined || x.symbol.contains(code_point));
            if (transition === undefined)
                break;
            state = transition.state;
            index += 1;
        }
        if (last_match === undefined)
            return false;
        if (options?.machine_id !== undefined) {
            if (last_match_state.machine_id !== options.machine_id)
                return false;
        }
        return symbols.slice(0, last_match).join('');
    }
}
exports.State = State;
//# sourceMappingURL=state.js.map