"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.RegularExpression = void 0;
const context_1 = require("./regex/context");
const single_symbol_1 = require("./regex/single-symbol");
const error_1 = require("./error");
const multi_symbol_1 = require("./regex/multi-symbol");
const integer_intervals_set_1 = require("./utils/integer-intervals-set");
const abstract_symbol_1 = require("./regex/abstract-symbol");
const source_reader_1 = require("./source/source-reader");
function code_point_to_printable(code_point) {
    return Number.isNaN(code_point) ? 'EOF' : String.fromCodePoint(code_point);
}
const DIGITS_INTERVAL = Object.freeze(new integer_intervals_set_1.IntegerInterval(0x30, 0x39 + 1));
const UPPERCASE_LETTERS_INTERVAL = Object.freeze(new integer_intervals_set_1.IntegerInterval(0x41, 0x5A + 1));
const LOWERCASE_LETTERS_INTERVAL = Object.freeze(new integer_intervals_set_1.IntegerInterval(0x61, 0x7A + 1));
const META_DIGITS_SYMBOL = new multi_symbol_1.MultiSymbol([DIGITS_INTERVAL]);
const META_ALPHANUMERIC_UNDERSCORE_SYMBOL = new multi_symbol_1.MultiSymbol([
    DIGITS_INTERVAL,
    UPPERCASE_LETTERS_INTERVAL,
    LOWERCASE_LETTERS_INTERVAL,
    0x5F
]);
const META_WHITESPACE_SYMBOL = new multi_symbol_1.MultiSymbol([
    0x20,
    0x09,
    0x0A,
    0x0B,
    0x0C,
    0x0D,
]);
const META_HORIZONTAL_TAB = new single_symbol_1.SingleSymbol(0x09);
const META_LINE_FEED = new single_symbol_1.SingleSymbol(0x0A);
const META_VERTICAL_TAB = new single_symbol_1.SingleSymbol(0x0B);
const META_FORM_FEED = new single_symbol_1.SingleSymbol(0x0C);
const META_CARRIAGE_RETURN = new single_symbol_1.SingleSymbol(0x0D);
class RegularExpression {
    constructor(reader, options) {
        this.reader = reader instanceof source_reader_1.SourceReader ? reader : new source_reader_1.SourceReader(reader);
        this.current_code_point = this.reader.read().codePointAt(0) || NaN;
        this.context = options?.context || new context_1.Context();
    }
    consume_current_code_point() {
        const code_point = this.current_code_point;
        this.current_code_point = this.reader.read().codePointAt(0) || NaN;
        return code_point;
    }
    expect_current_code_point_then_consume(expected_code_point) {
        if (this.current_code_point !== expected_code_point) {
            const current = code_point_to_printable(this.current_code_point);
            const expected = code_point_to_printable(expected_code_point);
            throw new error_1.CodeError(this.reader.file, this.reader.get_point(), `Expected \`${expected}\` character, but \`${current}\` found.`);
        }
        this.consume_current_code_point();
    }
    static is_valid_letter(code_point) {
        return META_ALPHANUMERIC_UNDERSCORE_SYMBOL.contains(code_point);
    }
    parse_symbol() {
        if (this.current_code_point === 0x2E) {
            this.consume_current_code_point();
            return new multi_symbol_1.MultiSymbol([new integer_intervals_set_1.IntegerInterval(0, Number.MAX_SAFE_INTEGER)]);
        }
        if (this.current_code_point === 0x5C) {
            this.consume_current_code_point();
            const character = this.consume_current_code_point();
            switch (character) {
                case 0x64: {
                    return META_DIGITS_SYMBOL;
                }
                case 0x77: {
                    return META_ALPHANUMERIC_UNDERSCORE_SYMBOL;
                }
                case 0x73: {
                    return META_WHITESPACE_SYMBOL;
                }
                case 0x74: {
                    return META_HORIZONTAL_TAB;
                }
                case 0x6E: {
                    return META_LINE_FEED;
                }
                case 0x76: {
                    return META_VERTICAL_TAB;
                }
                case 0x66: {
                    return META_FORM_FEED;
                }
                case 0x72: {
                    return META_CARRIAGE_RETURN;
                }
                default:
                    return new single_symbol_1.SingleSymbol(character);
            }
        }
        if (RegularExpression.is_valid_letter(this.current_code_point)) {
            const start = this.consume_current_code_point();
            if (this.current_code_point === 0x2D) {
                this.consume_current_code_point();
                const end = this.consume_current_code_point();
                if (!RegularExpression.is_valid_letter(end)) {
                    this.reader.revoke();
                    const found = code_point_to_printable(end);
                    throw new error_1.CodeError(this.reader.file, this.reader.get_point(), `Expecting last character of interval, but \`${found}\` found.`);
                }
                return new multi_symbol_1.MultiSymbol([new integer_intervals_set_1.IntegerInterval(start, end + 1)]);
            }
            else {
                return new single_symbol_1.SingleSymbol(start);
            }
        }
        return undefined;
    }
    parse_atom() {
        const symbol = this.parse_symbol();
        if (symbol === undefined)
            return undefined;
        const initial_state = this.context.create_new_state();
        const final_state = this.context.create_new_state();
        initial_state.add_transition(symbol, final_state);
        return { initial_state, final_state };
    }
    parse_brackets() {
        let negated = false;
        if (this.current_code_point === 0x5E) {
            this.consume_current_code_point();
            negated = true;
        }
        let symbol = new multi_symbol_1.MultiSymbol([]);
        while (true) {
            const s = this.parse_symbol();
            if (s === undefined)
                break;
            symbol = abstract_symbol_1.AbstractSymbol.merge(symbol, s);
        }
        if (negated)
            symbol = abstract_symbol_1.AbstractSymbol.negate(symbol);
        const initial_state = this.context.create_new_state();
        const final_state = this.context.create_new_state();
        initial_state.add_transition(symbol, final_state);
        return { initial_state, final_state };
    }
    parse_block() {
        if (this.current_code_point === 0x28) {
            this.consume_current_code_point();
            const sub_machine = this.parse_alternation();
            this.expect_current_code_point_then_consume(0x29);
            return sub_machine;
        }
        if (this.current_code_point === 0x5B) {
            this.consume_current_code_point();
            const sub_machine = this.parse_brackets();
            this.expect_current_code_point_then_consume(0x5D);
            return sub_machine;
        }
        return this.parse_atom();
    }
    parse_quantifier() {
        const block = this.parse_block();
        if (block === undefined)
            return undefined;
        if (this.current_code_point === 0x3F) {
            this.consume_current_code_point();
            const initial_state = this.context.create_new_state();
            const final_state = this.context.create_new_state();
            initial_state.add_epsilon_transition(block.initial_state);
            block.final_state.add_epsilon_transition(final_state);
            initial_state.add_epsilon_transition(final_state);
            return { initial_state, final_state };
        }
        if (this.current_code_point === 0x2A) {
            this.consume_current_code_point();
            const initial_state = this.context.create_new_state();
            const final_state = this.context.create_new_state();
            initial_state.add_epsilon_transition(block.initial_state);
            block.final_state.add_epsilon_transition(initial_state);
            initial_state.add_epsilon_transition(final_state);
            return { initial_state, final_state };
        }
        if (this.current_code_point === 0x2B) {
            this.consume_current_code_point();
            const initial_state = this.context.create_new_state();
            const final_state = this.context.create_new_state();
            initial_state.add_epsilon_transition(block.initial_state);
            block.final_state.add_epsilon_transition(block.initial_state);
            block.final_state.add_epsilon_transition(final_state);
            return { initial_state, final_state };
        }
        return block;
    }
    parse_concatenation() {
        const initial_state = this.context.create_new_state();
        const final_state = this.context.create_new_state();
        let last_state = initial_state;
        while (true) {
            const quantifier = this.parse_quantifier();
            if (quantifier === undefined)
                break;
            last_state.add_epsilon_transition(quantifier.initial_state);
            last_state = quantifier.final_state;
        }
        last_state.add_epsilon_transition(final_state);
        return { initial_state, final_state };
    }
    parse_alternation() {
        const initial_state = this.context.create_new_state();
        const final_state = this.context.create_new_state();
        while (true) {
            const alternation = this.parse_concatenation();
            initial_state.add_epsilon_transition(alternation.initial_state);
            alternation.final_state.add_epsilon_transition(final_state);
            if (this.current_code_point !== 0x7C)
                break;
            this.consume_current_code_point();
        }
        return { initial_state, final_state };
    }
    generate() {
        const machine = this.parse_alternation();
        machine.final_state.is_final = true;
        machine.initial_state.remove_non_determinism(this.context);
        return machine.initial_state;
    }
    static merge(context, state_machines) {
        const initial_state = context.create_new_state();
        const seen_states = new Set();
        state_machines.forEach(state => {
            state.get_transitively_reachable_final_states().forEach((final_state) => {
                if (seen_states.has(final_state.id))
                    throw new Error(`Provided states have non-disjunctive final states.`);
                final_state.machine_id = state.machine_id;
                seen_states.add(final_state.id);
            });
            initial_state.add_epsilon_transition(state);
        });
        initial_state.remove_non_determinism(context);
        return initial_state;
    }
}
exports.RegularExpression = RegularExpression;
//# sourceMappingURL=regular-expression.js.map