"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.analyze = void 0;
const declaration_1 = require("./ast/declaration");
const production_node_1 = require("./ast/production-node");
const error_1 = require("./error");
function starts_with_lowercase(name) {
    const first = name.charAt(0);
    return first === first.toLowerCase();
}
function analyze(file, source) {
    const terminals = new Map();
    const variables = new Map();
    let delimiter;
    for (const declaration of source.declarations) {
        const name = declaration.name;
        switch (declaration.type) {
            case declaration_1.DeclarationType.DELIMITER: {
                if (delimiter !== undefined) {
                    throw new error_1.CodeError(file, declaration.location, `Duplicate delimiter declaration detected. Firstly declared at ${delimiter.location}.`);
                }
                else {
                    delimiter = declaration;
                }
                break;
            }
            case declaration_1.DeclarationType.TERMINAL: {
                if (!starts_with_lowercase(name))
                    throw new error_1.CodeError(file, declaration.location, `Declared terminal \`${name}\` have to start with a lower-case letter.`);
                if (terminals.has(name)) {
                    const other = terminals.get(name);
                    throw new error_1.CodeError(file, declaration.location, `Duplicate terminal \`${name}\` declaration detected. Firstly declared at ${other.location}.`);
                }
                terminals.set(name, declaration);
                break;
            }
            case declaration_1.DeclarationType.VARIABLE: {
                if (starts_with_lowercase(name))
                    throw new error_1.CodeError(file, declaration.location, `Declared variable \`${name}\` have to start with an upper-case letter.`);
                if (variables.has(name)) {
                    const other = variables.get(name);
                    throw new error_1.CodeError(file, declaration.location, `Duplicate variable \`${name}\` declaration detected. Firstly declared at ${other.location}.`);
                }
                variables.set(name, declaration);
                break;
            }
            default:
                throw new Error('Lack of implementation.');
        }
    }
    if (variables.size === 0) {
        throw new error_1.CodeError(file, source.location, `No variables declared. At least one required.`);
    }
    for (const variable of variables.values()) {
        if (variable.productions.length === 0) {
            throw new error_1.CodeError(file, variable.location, `Variable \`${variable.name}\` has no productions. At least one required.`);
        }
    }
    for (const variable of variables.values()) {
        for (const production of variable.productions) {
            for (const node of production.nodes) {
                switch (node.type) {
                    case production_node_1.ProductionNodeType.TERMINAL_USAGE: {
                        const terminal = terminals.get(node.name);
                        if (terminal === undefined) {
                            throw new error_1.CodeError(file, node.location, `Using undeclared terminal \`${node.name}\`.`);
                        }
                        node.reference = terminal;
                        break;
                    }
                    case production_node_1.ProductionNodeType.VARIABLE_USAGE: {
                        const variable = variables.get(node.name);
                        if (variable === undefined) {
                            throw new error_1.CodeError(file, node.location, `Using undeclared variable \`${node.name}\`.`);
                        }
                        node.reference = variable;
                        break;
                    }
                    default:
                        throw new Error('Lack of implementation.');
                }
            }
        }
    }
    left_recursion_check_for: for (const variable of variables.values()) {
        const seen = new Set();
        const resolutions = [{ current: variable, previous: [] }];
        while (true) {
            const cursor = resolutions.shift();
            if (cursor === undefined)
                break left_recursion_check_for;
            if (seen.has(cursor.current.name)) {
                const names = cursor.previous.map(x => x.name);
                while (names[0] !== cursor.current.name)
                    names.shift();
                const path = `${names.join('\u2192')}\u2192${cursor.current.name}`;
                throw new error_1.CodeError(file, variable.location, `Left-recursion loop detected: ${path}`);
            }
            seen.add(cursor.current.name);
            for (const production of cursor.current.productions) {
                for (const node of production.nodes) {
                    if (node.type === production_node_1.ProductionNodeType.TERMINAL_USAGE)
                        break;
                    if (node.type === production_node_1.ProductionNodeType.VARIABLE_USAGE) {
                        const target = node.reference;
                        resolutions.push({ current: target, previous: [...cursor.previous, cursor.current] });
                        if (!target.epsilon)
                            break;
                    }
                }
            }
        }
    }
    for (const variable of variables.values()) {
        const used_terminals = new Map();
        for (const terminals_map of variable.terminals_maps) {
            for (const terminal of terminals_map.terminals) {
                const used = used_terminals.get(terminal.name);
                if (used !== undefined) {
                    throw new error_1.CodeError(file, variable.location, `Variable \`${variable.name}\` is ambiguous, terminal \`${terminal.name}\` leads to multiple productions.`);
                }
                used_terminals.set(terminal.name, terminal);
            }
        }
    }
}
exports.analyze = analyze;
//# sourceMappingURL=semantic-analysis.js.map