"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.parse = void 0;
const lexical_analysis_1 = require("./lexical-analysis");
const source_1 = require("./ast/source");
const variable_1 = require("./ast/variable");
const error_1 = require("./error");
const location_1 = require("./source/location");
const production_1 = require("./ast/production");
const variable_usage_1 = require("./ast/variable-usage");
const terminal_usage_1 = require("./ast/terminal-usage");
const terminal_1 = require("./ast/terminal");
const regular_expression_1 = require("./regular-expression");
const reader_1 = require("./reader");
const source_reader_1 = require("./source/source-reader");
const delimiter_1 = require("./ast/delimiter");
class SyntaxParser {
    constructor(regex_context, file, symbols) {
        this.regex_context = regex_context;
        this.file = file;
        this.symbols = symbols;
        this.index = 0;
    }
    get current_symbol() {
        return this.symbols[this.index];
    }
    consume() {
        const symbol = this.current_symbol;
        this.index += 1;
        return symbol;
    }
    require(expected) {
        const symbol = this.current_symbol;
        if (symbol.type !== expected) {
            const message = `expected token \`${lexical_analysis_1.SymbolType[expected]}\`, but found \`${lexical_analysis_1.SymbolType[symbol.type]}\``;
            throw new error_1.CodeError(this.file, symbol.location, message);
        }
        return this.consume();
    }
    collect(f, options) {
        const collection = [];
        while (true) {
            const element = f.bind(this)();
            if (element === undefined) {
                if (options?.at_least_one !== undefined && collection.length === 0)
                    throw new error_1.CodeError(this.file, this.current_symbol.location, `Expected at least one \`${options.at_least_one.node_name}\`, but zero found.`);
                return collection;
            }
            collection.push(element);
        }
    }
    parse_production_node() {
        switch (this.current_symbol.type) {
            case lexical_analysis_1.SymbolType.IDENTIFIER: {
                const id = this.consume();
                const first = id.lexeme.charAt(0);
                return (first === first.toUpperCase())
                    ? new variable_usage_1.VariableUsage(id.location.get_location(), id.lexeme)
                    : new terminal_usage_1.TerminalUsage(id.location.get_location(), id.lexeme);
            }
            default:
                return undefined;
        }
    }
    parse_production() {
        switch (this.current_symbol.type) {
            case lexical_analysis_1.SymbolType.PRODUCTION: {
                const start = this.consume().location.get_location().start;
                const nodes = this.collect(this.parse_production_node, { at_least_one: { node_name: 'PRODUCTION_NODE' } });
                const end = nodes[nodes.length - 1].location.get_location().end;
                return new production_1.Production(new location_1.Location(start, end), nodes);
            }
            case lexical_analysis_1.SymbolType.EPSILON: {
                const location = this.consume().location;
                return production_1.Production.create_epsilon(location);
            }
            default: {
                return undefined;
            }
        }
    }
    parse_declarations() {
        switch (this.current_symbol.type) {
            case lexical_analysis_1.SymbolType.DELIMITER: {
                const start = this.consume().location.get_location().start;
                const regex_symbol = this.require(lexical_analysis_1.SymbolType.REGEX);
                const end = regex_symbol.location.get_location().end;
                const reader = new source_reader_1.SourceReader(new reader_1.StringReader(regex_symbol.lexeme.slice(1, -1)), {
                    file: this.file,
                    location_offset: start
                });
                const regex = (new regular_expression_1.RegularExpression(reader, { context: this.regex_context })).generate();
                return new delimiter_1.Delimiter(new location_1.Location(start, end), regex);
            }
            case lexical_analysis_1.SymbolType.TERMINAL: {
                const start = this.consume().location.get_location().start;
                const name_symbol = this.require(lexical_analysis_1.SymbolType.IDENTIFIER);
                let regex_source = name_symbol.lexeme;
                let end = name_symbol.location;
                if (this.current_symbol.type === lexical_analysis_1.SymbolType.REGEX) {
                    const regex_symbol = this.consume();
                    regex_source = regex_symbol.lexeme.slice(1, -1);
                    end = regex_symbol.location;
                }
                const reader = new source_reader_1.SourceReader(new reader_1.StringReader(regex_source), {
                    file: this.file,
                    location_offset: start
                });
                const regex = (new regular_expression_1.RegularExpression(reader, { context: this.regex_context })).generate();
                return new terminal_1.Terminal(new location_1.Location(start, end.get_location().end), name_symbol.lexeme, regex);
            }
            case lexical_analysis_1.SymbolType.VARIABLE: {
                const start = this.consume().location.get_location().start;
                const name_symbol = this.require(lexical_analysis_1.SymbolType.IDENTIFIER);
                this.require(lexical_analysis_1.SymbolType.BRACES_LEFT);
                const productions = this.collect(this.parse_production);
                const end = this.require(lexical_analysis_1.SymbolType.BRACES_RIGHT).location.get_location().end;
                return new variable_1.Variable(new location_1.Location(start, end), name_symbol.lexeme, productions);
            }
            default: {
                return undefined;
            }
        }
    }
    parse() {
        const declarations = this.collect(this.parse_declarations);
        const eof_symbol = this.require(lexical_analysis_1.SymbolType.EOF);
        const location = new location_1.Location(new location_1.Point(1, 1), eof_symbol.location.get_location().end);
        return new source_1.Source(location, declarations);
    }
}
function parse(regex_context, file, symbols) {
    return (new SyntaxParser(regex_context, file, symbols)).parse();
}
exports.parse = parse;
//# sourceMappingURL=syntax-analysis.js.map