"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.IntegerIntervalsSet = exports.IntegerInterval = void 0;
class IntegerInterval {
    constructor(start, end) {
        this.start = start;
        this.end = end;
        if (!Number.isInteger(start) || !Number.isInteger(end))
            throw new Error('Integer interval must have integer boundaries.');
        if (this.length < 0)
            throw new Error('Integer interval must have non-negative length.');
    }
    get length() {
        return this.end - this.start;
    }
    contains(x) {
        return x >= this.start && x < this.end;
    }
}
exports.IntegerInterval = IntegerInterval;
class IntegerIntervalsSet {
    constructor() {
        this.intervals = [];
    }
    find_number_interval(x) {
        let left = 0;
        let right = this.intervals.length;
        while (true) {
            const length = right - left;
            if (length === 0)
                return { index: left, present: false };
            const pivot = left + Math.floor(length / 2);
            const interval = this.intervals[pivot];
            if (interval.contains(x))
                return { index: pivot, present: true };
            if (x < interval.start) {
                right = pivot;
            }
            else {
                left = pivot + 1;
            }
        }
    }
    add(interval) {
        let [start, end] = (typeof interval === 'number')
            ? [interval, interval + 1]
            : [interval.start, interval.end];
        const r0 = this.find_number_interval(start);
        const r1 = this.find_number_interval(end);
        let insert_index = r0.index;
        let to_remove = r1.index - r0.index;
        if (r0.present) {
            start = this.intervals[r0.index].start;
        }
        else {
            const left_index = r0.index - 1;
            if (left_index >= 0 && start === this.intervals[left_index].end) {
                start = this.intervals[left_index].start;
                insert_index = left_index;
                to_remove += 1;
            }
        }
        if (r1.present) {
            end = this.intervals[r1.index].end;
            to_remove += 1;
        }
        else {
            if (r1.index < this.intervals.length && end === this.intervals[r1.index].start) {
                end = this.intervals[r1.index].end;
                to_remove += 1;
            }
        }
        this.intervals.splice(insert_index, to_remove, new IntegerInterval(start, end));
    }
    get size() {
        return this.intervals.reduce((sum, interval) => sum + interval.length, 0);
    }
    get capacity() {
        return this.intervals.length;
    }
    contains(x) {
        const result = this.find_number_interval(x);
        return result.present;
    }
    to_array() {
        const numbers = [];
        for (let interval of this.intervals) {
            for (let i = interval.start; i < interval.end; i += 1)
                numbers.push(i);
        }
        return numbers;
    }
    get_intervals() {
        return this.intervals;
    }
    static calculate_union(left, right) {
        const set = new IntegerIntervalsSet();
        set.intervals = [...left.intervals];
        for (let interval of right.intervals)
            set.add(interval);
        return set;
    }
    static calculate_negation(set) {
        const negation = new IntegerIntervalsSet();
        let start = Number.MIN_SAFE_INTEGER;
        for (let interval of set.intervals) {
            if (start < interval.start)
                negation.add(new IntegerInterval(start, interval.start));
            start = interval.end;
        }
        if (start < Number.MAX_SAFE_INTEGER)
            negation.add(new IntegerInterval(start, Number.MAX_SAFE_INTEGER));
        return negation;
    }
    static calculate_differences_and_intersection(left, right) {
        const left_difference = new IntegerIntervalsSet();
        const intersection = new IntegerIntervalsSet();
        const right_difference = new IntegerIntervalsSet();
        const x = [
            { index: 0, start: undefined, intervals: left.intervals, difference: left_difference },
            { index: 0, start: undefined, intervals: right.intervals, difference: right_difference },
        ];
        while (true) {
            for (let key = 0; key < 2; ++key) {
                if (x[key].index >= x[key].intervals.length) {
                    const other = x[key ^ 1];
                    for (let i = other.index; i < other.intervals.length; ++i) {
                        if (other.start === undefined) {
                            other.difference.intervals.push(other.intervals[i]);
                        }
                        else {
                            const end = other.intervals[i].end;
                            const interval = new IntegerInterval(other.start, end);
                            other.difference.intervals.push(interval);
                            other.start = undefined;
                        }
                    }
                    return { left_difference, intersection, right_difference };
                }
            }
            if (x[0].start === undefined)
                x[0].start = x[0].intervals[x[0].index].start;
            if (x[1].start === undefined)
                x[1].start = x[1].intervals[x[1].index].start;
            let start_difference = x[0].start - x[1].start;
            if (start_difference === 0) {
                const left_end = x[0].intervals[x[0].index].end;
                const end_difference = left_end - x[1].intervals[x[1].index].end;
                if (end_difference === 0) {
                    intersection.intervals.push(new IntegerInterval(x[0].start, left_end));
                    x[0].index += 1;
                    x[1].index += 1;
                    x[0].start = undefined;
                    x[1].start = undefined;
                }
                else {
                    const nearest_index = (end_difference < 0) ? 0 : 1;
                    const nearest = x[nearest_index];
                    const nearest_interval = nearest.intervals[nearest.index];
                    const farthest = x[nearest_index ^ 1];
                    intersection.intervals.push(new IntegerInterval(nearest.start, nearest_interval.end));
                    nearest.index += 1;
                    nearest.start = undefined;
                    farthest.start = nearest_interval.end;
                }
            }
            else {
                const nearest_index = (start_difference < 0) ? 0 : 1;
                const nearest = x[nearest_index];
                const nearest_interval = nearest.intervals[nearest.index];
                const farthest = x[nearest_index ^ 1];
                if (nearest_interval.end <= farthest.start) {
                    nearest.difference.intervals.push(new IntegerInterval(nearest.start, nearest_interval.end));
                    nearest.index += 1;
                    nearest.start = undefined;
                }
                else {
                    nearest.difference.intervals.push(new IntegerInterval(nearest.start, farthest.start));
                    nearest.start = farthest.start;
                }
            }
        }
    }
}
exports.IntegerIntervalsSet = IntegerIntervalsSet;
//# sourceMappingURL=integer-intervals-set.js.map